#%%
import os
import shutil
import numpy as np
import casadi as ca
from utils import rk4step
from newton_method import newton_method
from plot_utils import plotTraj

#%%
nonlinear_dyn = False            # set to True for nonlinear dynamics
control_penalty_steep = False    # set to True for steep control penalty (a^4)

# set to True if all plots should be merged into one pdf
# needs PyPDF2 package installed (pip install PyPDF2)
merge_pdfs = False      
if merge_pdfs:
    from PyPDF2 import PdfMerger

# for latex fonts in plot. needs latex installed.
use_latex_fonts = False
if use_latex_fonts:
    from latexify import latexify
    latexify(fig_width=3.5, fig_height=3)

# folder for saving results
outfolder = 'results/'
if nonlinear_dyn:
    outfolder += 'nl_dyn'
else:
    outfolder += 'lin_dyn'
if control_penalty_steep:
    outfolder += '_steep_contr_pen'
if os.path.exists(outfolder):
    shutil.rmtree(outfolder, ignore_errors=False, onerror=None)
os.makedirs(outfolder)

# problem parameters
T = 4               # continous time horizon
N = 30              # discrete time horizon
dt = T/N            # time step / control discretization
N_int = 10          # number of integrator steps per time step
ns = 2              # number of states
na = 1              # number of actions
k_spring = 1        # spring constant
a_grav = 1          # gravity acceleration
L = 1               # pendulum length

s0bar = np.array([2, 0])        # initial state
sNbar = np.array([0, 0])        # target state

# define ode
s = ca.SX.sym('x', ns)          # symbol for state
a = ca.SX.sym('u', na)          # symbol for action
p = s[0]                        # position
v = s[1]                        # velocity

pdot = v                        # time derivative of position is velocity
if nonlinear_dyn:
    vdot = a_grav * ca.sin((a-p / L))
else:
    vdot = k_spring * (a - p)

sdot = ca.veccat(pdot, vdot )               # time derivative of state / ode
ode = ca.Function('ode', [s, a], [ sdot ])  # as function

# define integrator of dynamics
# Runge-Kutta 4th order integrator
s_next = rk4step(ode, dt/N_int, s, a)                   # as casadi expression
rk4_step = ca.Function('rk4_step', [s, a], [s_next])    # as casadi function
# apply function rk4_step() N_int times on itself
integrator = rk4_step.mapaccum(N_int)       

# define NLP
# decision variables
a_all = ca.SX.sym('a_all', na, N)       # all controls
s_all = ca.SX.sym('s_all', ns, N+1)     # all states
x = ca.veccat(a_all, s_all)             # all decision variables together
nx = x.shape[0]                         # decision variable dimension

# objective function
obj = 0
for i in range(N):
    obj += a_all[:,i]**2
    obj += .01 * s_all[0,i]**2
if control_penalty_steep:
    for i in range(N):
        obj += a_all[:,i]**4

# constraints
constr = []
constr.append(s_all[:,0] - s0bar)           # initial condition
for i in range(N):                          # dynamics        
    constr.append(s_all[:,i+1] - integrator(s_all[:,i], a_all[:,i])[:,-1])
constr.append(s_all[:,N] - sNbar)           # terminal constraint

c = ca.veccat(*constr)                  # all equality constraints together
nc = c.shape[0]                         # number of (scalar) equality constraints

lam = ca.SX.sym('lam', nc)              # lagrange multiplier of equality constraints
Lagrangian = obj - lam.T @ c            # Lagrangian
nab_Lag = ca.gradient(Lagrangian, x)    # gradient of Lagrangian

# root finding problem (F(z) = 0)
z = ca.vertcat(x, lam)                  # all variables
F = ca.Function('F', [z], [ca.vertcat(nab_Lag, c)])

# solve root finding problem using newtons method
iters = newton_method(F, np.zeros(nx+nc))

# go through all iterations and visualize (saves plots to pdfs)
filenames = []
for i, iter in enumerate(iters):
    X_ = ca.evalf(ca.substitute(s_all, z, iter)).full()
    U_ = ca.evalf(ca.substitute(a_all, z, iter)).full()
    filename = outfolder + '/iter_{}.pdf'.format(i)
    filenames.append(filename)
    if not nonlinear_dyn and not control_penalty_steep:
        title = None
    else:
        title = 'Iteration {}'.format(i)
    plotTraj(X_, U_, T, filename=filename, integrator=integrator, title=title)

if merge_pdfs:
    print('Concatenating all pdfs...')
    merger = PdfMerger()
    for pdf in filenames:
        merger.append(pdf)
    merger.write(outfolder +'/iter_all.pdf')
    merger.close()
    print('Done.')